---------------------------------------------------------------
--                                                           
--  PENCIL AND PAPER INTERPRETER
--  By: Martin Carlisle and Al White     
--                                                           
--  PENCIL_AND_PAPER_COMPUTER.ADB 
--  Description : see below 
--                       
--  By: Al White
--                                         
-- PENCIL AND PAPER INTERPRETER is free software; you can redistribute 
-- it and/or modify it without restriction.  However, we ask that you
-- please retain the original author information, and clearly  
-- indicate if it has been modified.
--  
-- PENCIL AND PAPER INTERPRETER is distributed in the hope that it will 
-- be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
--
---------------------------------------------------------------
-- Overview: The procedure (Reset_Program_counter) will
-- set the Program_Counter to 0, clear the accumulator, and start the program by 
-- setting program_running to true
--
-- The procedure do_one_instruction is used to fetch and execute
-- each instruction until either a STOP instruction
-- is encountered, the program counter reaches the end of memory, or an exception is raised. 
--
--------------------------------------------------------------------------------------------- 
with Ada.Text_Io, Ada.Integer_Text_Io, Ada.Unchecked_Conversion;
use Ada.Text_Io, Ada.Integer_Text_Io;
package body Pencil_And_Paper_Computer is

   -- Memory is implemented as an array of bytes using big endian bit order. 
   Max_Memory : constant := 31;    --32 bytes                                    
   subtype Bit is Integer range 0..1;
   type Byte_Type is array (0 .. 7) of Bit; 
   pragma Pack(Byte_Type);
   type Memory_Type is array (0 .. Max_Memory) of Byte_Type; 
   package Instruction_Io is new Ada.Text_Io.Enumeration_Io(Enum => Instruction);


   -- Global Variable Declarations
   Memory          : Memory_Type;  
   Program_Counter : Memory_Location;  
   Accumulator     : Accumulator_Value;  
   Program_Running : Boolean           := False;  

   ------------------------------------------------------------------------
   -- PRIVATE subprogram declarations  
   --
   -- The following functions and procedures are used by the
   -- subprograms listed in the specification and are not exposed by the spec
   ------------------------------------------------------------------------
   --
   -- FUNCTION INT2BIN
   -- Int2Bin converts a number of type memory_value into the 8-bit byte_type
   function Int2bin is 
   new Ada.Unchecked_Conversion(Memory_Value, Byte_Type);


   -- FUNCTION BIN2INT
   -- bin2int converts a number of type byte_type into memory_value
   function Bin2int is 
   new Ada.Unchecked_Conversion(Byte_Type, Memory_Value);


   --PROCEDURE UNASSEMBLE
   -- This procedure is used to, given the value stored in a memory address, return the 
   -- opcode of type instruction and operand of type integer
   procedure Unassemble (
         Rawvalue : in     Memory_Value; 
         Opcode   :    out Instruction;  
         Operand  :    out Integer       ) is 
   begin
      Opcode := Instruction'Val(Rawvalue/2**5);
      Operand := Rawvalue mod 2**5;
   end Unassemble;


   ----------------------------------------------------------
   -- PUBLIC subprogram declarations
   --
   -- These functions and procedures are exposed by the spec
   ----------------------------------------------------------

   -- PROCEDURE SET_LOCATION
   -- This procedure stores a value (memory_value as integer) in the memory_location (also an integer)
   -- except the memory_value is converted to byte_type before storage
   procedure Set_Location (
         Address : in     Memory_Location; 
         Value   : in     Memory_Value     ) is 
   begin
      Memory(Address):=Int2bin(Value);
   end Set_Location;


   --FUNCTION GET_LOCATION
   -- This function will return the integer representation of a memory value
   function Get_Location (
         Address : in     Memory_Location ) 
     return Memory_Value is 
   begin
      return Bin2int(Memory(Address));
   end Get_Location;

   --FUNCTION GET_LOCATION
   -- This function will return the string representation of the opcode and operand at a given
   -- memory location
   function Get_Location (
         Address : in     Memory_Location ) 
     return String is 
      Opcode        : Instruction;  
      Operand       : Integer;  
      String_Output : String (1 .. 8) := (others => ' ');       -- a 2 char length string representation of the operand            
   begin
      Unassemble(Bin2int(Memory(Address)), Opcode, Operand);
      -- unassemble decodes value at memory address into 
      -- corresponding opcode/operand int values and is
      -- implemented in private section

      -- put together the string value of the opcode and string value of the operand into an 8-char string 

      Instruction_Io.Put(
         Item => Opcode,       
         To   => String_Output);
      Put(
         Item => Operand,               
         To   => String_Output (7 .. 8));
      return String_Output;
   end Get_Location;

   -- FUNCTION GET_ACCUMULATOR
   -- This function returns the value of the accumulator
   function Get_Accumulator return Accumulator_Value is 
   begin
      return Accumulator;
   end Get_Accumulator;

   procedure Set_Accumulator(Value : in Accumulator_Value) is 
   begin
      Accumulator := Value;
   end Set_Accumulator;

   -- FUNCTION GET_PROGRAM_COUNTER
   --This function returns the value of the program counter
   function Get_Program_Counter return Memory_Location is 
   begin
      return Program_Counter;
   end Get_Program_Counter;

   procedure Set_Program_Counter(Address : in Memory_Location) is 
   begin
      Program_Counter := Address;
   end Set_Program_Counter;

   -- PROCEDURE RESET_PROGRAM_COUNTER
   -- this procedure starts/restarts the program      
   procedure Reset_Program_Counter is 
   begin
      Program_Counter:=0;   --PC set to 0
      Accumulator:=0;       --Clear the accumulator
      Program_Running:=True;
   end Reset_Program_Counter;

   -- PROCEDURE POST_READ
   Remember_Memory_Location : Memory_Location;
   Remember_Update_Callback : Update_Callback_Type; 
   procedure post_read(readvalue: in memory_value) is
   begin
      Memory(Remember_Memory_Location):=Int2bin(Readvalue);
      Program_Counter:=Program_Counter+1;
      Remember_Update_Callback.all(Remember_Memory_Location);
   end post_read;

   -- PROCEDURE DO_ONE_INSTRUCTION
   -- This is the beef of the program that is called to fetch and execute each step of the program
   --
   procedure Do_One_Instruction (
         Update_Callback : in     Update_Callback_Type; 
         Read_Callback   : in     Read_Callback_Type;   
         Print_Callback  : in     Print_Callback_Type   ) is 
      Opcode    : Instruction;  
      Operand   : Integer;  
      post_read_callback : post_read_callback_type := post_read'access;   
   begin
      if Program_Running and Program_Counter<=31 then
         Unassemble(Bin2int(Memory(Program_Counter)), Opcode, Operand);
         --unassemble the value stored at the current proram-counter location in memory

         -- depending on the opcode, execute the instruction

         case Opcode is
            when Stop =>
               Program_Running:=False;
               raise Program_Halted;
            when Add =>
               -- add the contents at the address in the operand to the accumulator
               begin
                  Accumulator:=Accumulator+Bin2int(Memory(Operand));
               exception when Constraint_Error =>
                  raise Memory_Overflow;
               end;
               Program_Counter:=Program_Counter+1;
            when Sub =>
               -- subtract the contents at the address in the operand from the accumulator
               -- if subtraction operation would cause accumulator to go negative then store a 0
               if Accumulator-Bin2int(Memory(Operand))<0 then
                  Accumulator:=0;
                  Program_Counter:=Program_Counter+1;
                  raise Memory_Underflow;
               else
                  Accumulator:=Accumulator-Bin2int(Memory(Operand));
                  Program_Counter:=Program_Counter+1;
               end if;
            when Load =>
               -- place value at the address in operand into the accumulator
               Accumulator:=Bin2int(Memory(Operand));
               Program_Counter:=Program_Counter+1;
            when Store =>
               -- place copy of accumulator in memory referenced by operand
               if Accumulator > 255 then
                  Memory(Operand):=Int2bin(Accumulator mod 255);
                  Program_Counter:=Program_Counter+1;
                  Update_Callback.all(Operand);
                  raise Memory_Overflow;
               else

                  Memory(Operand):=Int2bin(Accumulator);
                  Program_Counter:=Program_Counter+1;
                  Update_Callback.all(Operand);
                  -- call update_callback when memory changes
               end if;
            when Read =>
               Remember_Update_Callback := Update_Callback;
               Remember_Memory_Location := Operand;
               Read_Callback.all(post_read_callback);
               -- get value from keyboard and store in memory ref by operand

            when Print =>
               -- print vale at memory location ref by operand
               Print_Callback.all(Bin2int(Memory(Operand)));
               Program_Counter:=Program_Counter+1;
            when Pjump =>
               -- jump to address ref by operand if accumulator is > 0
               if Accumulator > 0 then
                  Program_Counter:=Operand;
               else
                  Program_Counter:=Program_Counter+1;
               end if;
         end case;
         null;
      else
         Program_Running:=False;
         raise Program_Halted;
         --tried to do an instruction when program hadn't been started or tried to execute/jump outside of memory
      end if;

   end Do_One_Instruction;


   -- PROCEDURE READ_PROGRAM
   -- This procedure will read the contents of the file into memory
   -- if the program was not created in the correct format by WRITE_PROGRAM then
   -- unknown results could occur
   procedure Read_Program (
         Filename : in     String ) is 
      Input_File     : File_Type;  
      Memory_Counter,  
      Memory_Item    : Integer   := 0;  
   begin
      Open(Input_File, In_File, Filename);
      while not End_Of_File (Input_File) and Memory_Counter <= 31 loop
         Get (Input_File, Memory_Item);
         Set_Location(Memory_Counter, Memory_Item);
         Memory_Counter:=Memory_Counter+1;
      end loop;
      Close(Input_File);
   end Read_Program;

   -- PROCEDURE WRITE_PROGRAM
   -- this procedure will create a file and write the contents of memory (represented as an integer) 
   -- with each value on its own line
   procedure Write_Program (
         Filename : in     String ) is 
      Output_File : File_Type;  
   begin
      Create(Output_File, Name => Filename);
      for I in 0..31 loop
         Put(Output_File, Bin2int(Memory(I)));
         New_Line(Output_File);
      end loop;
      Close(Output_File);
   end Write_Program;

   -- returns first occasion of two stops in a row.  If there aren't
   -- two stops found, returns Memory_Location'Last-1
   function Find_Two_Stops return Natural is
      Stop_Count : Natural := 0;
      Opcode    : Instruction;  
      Operand   : Integer;  
   begin
      for I in Memory_Location loop
         Unassemble(Bin2int(Memory(I)), Opcode, Operand);
         if Opcode=stop then
            if Stop_Count = 0 then
               Stop_Count := 1;
            else
               return I-1;
            end if;
         else
            Stop_Count := 0;
         end if;
      end loop;
      return Memory_Location'Last-1;
   end Find_Two_Stops;
   

   procedure Insert_Instruction is
      PC : Natural;
      Two_Stops : Natural := Find_Two_Stops;
      Opcode    : Instruction;  
      Operand   : Integer;  
   begin
      PC := Get_Program_Counter;
      for I in 0..Two_Stops loop
         Unassemble(Bin2int(Memory(I)), Opcode, Operand);
         case Opcode is
            when Stop =>
               null;
            when Add|Sub|Load|Store|Read|Print|Pjump =>
               if Operand >= PC and Operand <= Two_Stops then
                  Memory(I) := Int2Bin(Bin2Int(Memory(I))+1);
               end if;
         end case;
      end loop;
      for I in reverse PC..Two_Stops loop
         Set_Location(
            Address => I+1,
            Value   => Get_Location(I));
      end loop;
   end Insert_Instruction;
   
   procedure Delete_Instruction is
      PC : Natural;
      Two_Stops : Natural := Find_Two_Stops;
      Opcode    : Instruction;  
      Operand   : Integer;  
   begin
      PC := Get_Program_Counter;
      for I in 0..Two_Stops loop
         Unassemble(Bin2int(Memory(I)), Opcode, Operand);
         case Opcode is
            when Stop =>
               null;
            when Add|Sub|Load|Store|Read|Print|Pjump =>
               if Operand > PC and Operand <= Two_Stops then
                  Memory(I) := Int2Bin(Bin2Int(Memory(I))-1);
               end if;
         end case;
      end loop;
      for I in PC..Two_Stops loop
         Set_Location(
            Address => I,
            Value   => Get_Location(I+1));
      end loop;
      Set_Location(
         Address => Two_Stops+1,
         Value   => Stop_Instruction);
   end Delete_Instruction;
end Pencil_And_Paper_Computer;